(use-modules (hoot ffi)
             ((hoot hashtables)
              #:select (make-weak-key-hashtable
                        weak-key-hashtable-ref
                        weak-key-hashtable-set!))
             (ice-9 match)
             (srfi srfi-9)
             (srfi srfi-11))

(define-foreign current-document
  "document" "get"
  -> (ref extern))
(define-foreign document-body
  "document" "body"
  -> (ref extern))
(define-foreign get-element-by-id
  "document" "getElementById"
  (ref string) -> (ref null extern))
(define-foreign make-text-node
  "document" "createTextNode"
  (ref string) -> (ref extern))
(define-foreign make-element
  "document" "createElement"
  (ref string) -> (ref extern))
(define-foreign make-tree-walker
  "document" "createTreeWalker"
  (ref extern) -> (ref extern))

(define-foreign element-value
  "element" "value"
  (ref extern) -> (ref string))
(define-foreign set-element-value!
  "element" "setValue"
  (ref extern) (ref string) -> none)
(define-foreign %element-checked?
  "element" "checked"
  (ref extern) -> i32)
(define (element-checked? elem)
  (= (%element-checked? elem) 1))
(define-foreign set-element-checked!
  "element" "setChecked"
  (ref extern) i32 -> none)
(define-foreign add-event-listener!
  "element" "addEventListener"
  (ref extern) (ref string) (ref extern) -> none)
(define-foreign remove-event-listener!
  "element" "removeEventListener"
  (ref extern) (ref string) (ref extern) -> none)
(define-foreign append-child!
  "element" "appendChild"
  (ref extern) (ref extern) -> (ref extern))
(define-foreign remove!
  "element" "remove"
  (ref extern) -> none)
(define-foreign replace-with!
  "element" "replaceWith"
  (ref extern) (ref extern) -> none)
(define-foreign set-attribute!
  "element" "setAttribute"
  (ref extern) (ref string) (ref string) -> none)
(define-foreign remove-attribute!
  "element" "removeAttribute"
  (ref extern) (ref string) -> none)

(define-foreign event-target
  "event" "target"
  (ref extern) -> (ref extern))
(define-foreign keyboard-event-key
  "event" "keyboardKey"
  (ref extern) -> (ref string))

(define-foreign current-node
  "treeWalker" "currentNode"
  (ref extern) -> (ref null extern))
(define-foreign set-current-node!
  "treeWalker" "setCurrentNode"
  (ref extern) (ref null extern) -> none)
(define-foreign next-node!
  "treeWalker" "nextNode"
  (ref extern) -> (ref null extern))
(define-foreign first-child!
  "treeWalker" "firstChild"
  (ref extern) -> (ref null extern))
(define-foreign next-sibling!
  "treeWalker" "nextSibling"
  (ref extern) -> (ref null extern))

(define-foreign get-item
  "localStorage" "getItem"
  (ref string) -> (ref string))
(define-foreign set-item
  "localStorage" "setItem"
  (ref string) (ref string) -> none)

(define procedure->external/cached
  (let ((cache (make-weak-key-hashtable)))
    (lambda (proc)
      (or (weak-key-hashtable-ref cache proc)
          (let ((f (procedure->external proc)))
            (weak-key-hashtable-set! cache proc f)
            f)))))

(define (add-event-listener!/wrap elem name proc)
  (add-event-listener! elem name (procedure->external/cached proc)))
(define (remove-event-listener!/wrap elem name proc)
  (remove-event-listener! elem name (procedure->external/cached proc)))

(define (sxml->dom exp)
  (match exp
    ((? string? str)
     (make-text-node str))
    (((? symbol? tag) . body)
     (let ((elem (make-element (symbol->string tag))))
       (define (add-children children)
         (for-each (lambda (child)
                     (append-child! elem (sxml->dom child)))
                   children))
       (match body
         ((('@ . attrs) . children)
          (for-each (lambda (attr)
                      (match attr
                        (((? symbol? name) (? string? val))
                         (set-attribute! elem
                                         (symbol->string name)
                                         val))
                        (((? symbol? name) (? procedure? proc))
                         (let ((name* (symbol->string name)))
                           (add-event-listener!/wrap elem name* proc)))))
                    attrs)
          (add-children children))
         (children (add-children children)))
       elem))))

(define (virtual-dom-render root old new)
  (define (attrs+children exp)
    (match exp
      ((('@ . attrs) . children)
       (values attrs children))
      (children
       (values '() children))))
  (define (find-attr attrs name)
    (match attrs
      (() #f)
      ((attr . rest)
       (match attr
         ((name* val)
          (if (eq? name name*)
              val
              (find-attr rest name)))))))
  (define (update-attrs node old-attrs new-attrs)
    (for-each
     (lambda (attr)
       (match attr
         ((name val)
          (let ((name-str (symbol->string name)))
            (match (find-attr old-attrs name)
              ;; No existing attr/listener, add new one.
              (#f
               (match val
                 ((? string?)
                  (set-attribute! node name-str val))
                 ((? procedure?)
                  (add-event-listener!/wrap node name-str val))))
              ;; Replace old attr or listener with new.
              (old-val
               (match val
                 ((? string?)
                  (unless (equal? old-val val)
                    (set-attribute! node name-str val)))
                 ((? procedure?)
                  (unless (eq? old-val val)
                    (remove-event-listener!/wrap node name-str old-val)
                    (add-event-listener!/wrap node name-str val))))))))))
     new-attrs)
    ;; Delete old attrs that aren't in new.
    (for-each
     (lambda (attr)
       (match attr
         ((name val)
          (let ((name-str (symbol->string name)))
            (match (find-attr new-attrs name)
              (#f
               (match val
                 ((? string?)
                  (remove-attribute! node name-str))
                 ((? procedure?)
                  (remove-event-listener! node name-str val))))
              (_ #t))))))
     old-attrs))
  (let ((walker (make-tree-walker root)))
    (first-child! walker)
    (let loop ((parent root)
               (old old)
               (new new))
      (match old
        (#f
         ;; It's the first render, so clear out whatever might be
         ;; in the actual DOM and render the entire tree.  No
         ;; diffing necessary.
         (let loop ((node (current-node walker)))
           (unless (external-null? node)
             (let ((next (next-sibling! walker)))
               (remove! node)
               (loop next))))
         (append-child! parent (sxml->dom new)))
        ((? string?)
         ;; Replace text node with either a new text node if the
         ;; text has changed, or an element subtree if the text
         ;; has been replaced by an element.
         (unless (and (string? new) (string=? old new))
           (let ((new-node (sxml->dom new)))
             (replace-with! (current-node walker) new-node)
             (set-current-node! walker new-node))))
        (((? symbol? old-tag) . old-rest)
         (let-values (((old-attrs old-children)
                       (attrs+children old-rest)))
           (match new
             ((? string?)
              ;; Old node was an element, but the new node is a
              ;; string, so replace the element subtree with a
              ;; text node.
              (let ((new-text (make-text-node new)))
                (replace-with! (current-node walker) new-text)
                (set-current-node! walker new-text)))
             (((? symbol? new-tag) . new-rest)
              (let-values (((new-attrs new-children)
                            (attrs+children new-rest)))
                (cond
                 ;; The element tag is the same, so modify the
                 ;; inner contents of the element if necessary.
                 ((eq? old-tag new-tag)
                  (let ((parent (current-node walker)))
                    (update-attrs parent old-attrs new-attrs)
                    (first-child! walker)
                    (let child-loop ((old old-children)
                                     (new new-children))
                      (match old
                        (()
                         ;; The old child list is empty, so
                         ;; diffing stops here.  All remaining
                         ;; children in the new list are fresh
                         ;; elements that need to be added.
                         (for-each
                          (lambda (new)
                            (append-child! parent (sxml->dom new)))
                          new))
                        ((old-child . old-rest)
                         (match new
                           ;; The new child list is empty, so any
                           ;; remaining children in the old child
                           ;; list need to be removed, including
                           ;; the current one.
                           (()
                            (let rem-loop ((node (current-node walker)))
                              (unless (external-null? node)
                                (let ((next (next-sibling! walker)))
                                  (remove! node)
                                  (rem-loop next)))))
                           ;; Recursively diff old and new child
                           ;; elements.
                           ((new-child . new-rest)
                            (loop parent old-child new-child)
                            (next-sibling! walker)
                            (child-loop old-rest new-rest))))))
                    (set-current-node! walker parent)))
                 ;; New element tag is different than the old
                 ;; one, so replace the entire element subtree.
                 (else
                  (let ((node (sxml->dom new)))
                    (replace-with! (current-node walker)
                                   node)
                    (set-current-node! walker node)))))))))))))

(define-record-type <slide>
  (make-slide title? name content)
  slide?
  (title? slide-title?)
  (name slide-name)
  (content slide-content))

(define* (slide name content #:key title?)
  (make-slide title? name content))

(define (template slide)
  (match slide
    (($ <slide> title? name content)
     `(main (@ (class "slideshow"))
            (article (@ (class ,(if title? "slide title" "slide")))
                     (h1 ,name)
                     (section ,@content))))))

(define *slides* '())
(define *prev-slides* '())
(define (current-slide) (car *slides*))
(define (next-slide)
  (match *slides*
    (() (values))
    ((_) (values))
    ((slide . rest)
     (set! *prev-slides* (cons slide *prev-slides*))
     (set! *slides* rest))))
(define (prev-slide)
  (match *prev-slides*
    (() (values))
    ((slide . rest)
     (set! *slides* (cons slide *slides*))
     (set! *prev-slides* rest))))

(define *current-vdom* #f)
(define (render)
  (let ((new-vdom (template (current-slide))))
    (virtual-dom-render (document-body) *current-vdom* new-vdom)
    (set! *current-vdom* new-vdom)))

(define (on-key-up event)
  (match (keyboard-event-key event)
    ("ArrowLeft"
     (prev-slide)
     (render))
    ("ArrowRight"
     (next-slide)
     (render))
    (_ (values))))
(add-event-listener! (current-document) "keyup" (procedure->external on-key-up))

(define (init-slides! slides)
  (set! *slides* slides)
  (render))

(define* (img src #:optional (attrs '()))
  `(img (@ (src ,src) ,@attrs)))

(define (video src)
  `(video (@ (src ,src) (loop "") (autoplay "true"))))

(init-slides!
 (list
  (slide "Why Spritely is all-in on WebAssembly"
         '((p "David Thompson")
           (p "Spritely Institute")
           (p "Dweb Camp 2024"))
         #:title? #t)
  (slide "Intro"
         `((p "Hi, I’m Dave!")
           ,(img "static/spritely-institute-logo-sketchysketch-800px.png"
                 '((class "white-background")))
           (p "I am CTO at the Spritely Institute!"))
         #:title? #t)
  (slide "Intro"
         `(,(video "static/spritely-flask-3d-preview6.mp4")
           (p "Spritely is a 501(c)(3) non-profit")
           (p "We’re building tech that makes secure, distributed programming easy!")
           (p (em "Learn more at our next talk in Hackers Hall at 10:15")))
         #:title? #t)
  (slide "Gotta go fast!"
         '()
         #:title? #t)
  (slide "Problem: Deployment"
         '((ul
            (li "We need to make it " (em "easy")
                " to get our tech in the hands of users")
            (li "Small team; can’t build for each desktop/mobile OS")
            (li "Target the web instead"))))
  (slide "On JavaScript"
         '((ul
            (li "JavaScript the language is pretty " (em "okay") " these days")
            (li "JavaScript the toolchain is a mess!")
            (li "We " (em "do") " want the same language on back/frontends")
            (li "... but we don’t want to use JavaScript"
                (ul
                 (li "No, not TypeScript, either")))
            (li "What do we do?"))))
  (slide "Why not compile to JavaScript?"
         `((ul
            (li "Language features are reduced to what JS can express"
                (ul
                 (li "All numbers are floats")
                 (li "No tail recursion")
                 (li "etc.")))
            (li "Did I mention that the toolchain is a mess?"))))
  (slide "The npm mess"
         '((ul
            (li "Huge dependency graphs that resist distro packaging")
            (li "Tower of Babel.js")
            (li "Fine for large companies with devops teams")
            (li "Bad for small teams")
            (li "Bad for self-hosted apps"
                (ul
                 (li "Need to be " (em "easily maintained")
                     " by one person"))))))
  (slide "Solution: WebAssembly"
         `((div (@ (class "row"))
                (ul
                 (li "Bytecode for the web (and beyond)")
                 (li "Capability security model (Spritely likey)")
                 (li "W3C open standard (it’s not Java applets again)"))
                ,(img "static/webassembly-logo.svg"))))
  (slide "Solution: WebAssembly (continued)"
         `((div (@ (class "row"))
                (ul
                 (li "Good compilation target (unlike JS)")
                 (li "Recently standardized GC-managed heap objects"
                     (ul
                      (li (em "Great") " target for dynamic languages!"))))
                ,(img "static/webassembly-logo.svg"))))
  (slide "Can we really avoid JavaScript?"
         '((ul
            (li "Almost!")
            (li "Wasm programs are booted from JavaScript"
                (ul
                 (li "Not much code")
                 (li "No need for complex JavaScript toolchain"))))))
  (slide "So you’re gonna use emscripten, right?"
         '((p (strong "Nope!"))))
  (slide "The emscripten mess"
         '((ul
            (li "LLVM-based toolchain, hodepodge of tools"
                (ul
                 (li "binaryen, wabt, wasm-bindgen, etc.")))
            (li "Primarily for static languages (C, Rust, etc.)")
            (li "For Wasm linear memory, " (em "not") " Wasm GC")
            (li "Binaries tend to be huge")
            (li "Compiling an existing runtime to Wasm is unsatisfying"))))
  (slide "Let’s build a new toolchain!"
         '((p "Goals:")
           (ul
            (li "Target Wasm GC")
            (li "Minimal dependencies"
                (ul
                 (li "No node/npm")
                 (li "No emscripten")))
            (li "Easy to build web applications "
                (em "from source")))))
  (slide "Hoot: Spritely’s Wasm toolkit"
         `((div (@ (class "row"))
                (div
                 (ul
                  (li "Full general-purpose Wasm toolchain")
                  (li "Ahead-of-time, whole-program compiler for Scheme")
                  (li "Small program; small binary")
                  (li "Built on Guile")))
                ,(img "static/hoot.png" '((style "width: 50%;"))))))
  (slide "Full Wasm toolchain"
         '((p "Hoot is a lighter alternative to the emscripten toolchain")
           (p "Components:"
              (ul
               (li "Assembler (Wasm→WAT)")
               (li "Disassembler (WAT→Wasm)")
               (li "Linker")
               (li "Interpreter")
               (li "...and more!")))))
  (slide "Scheme compiler"
         `((p
            (div (@ (class "row"))
                 (ul
                  (li "AOT, whole-program, optimizing")
                  (li "All the goodies:"
                      (ul
                       (li "Multi-value return")
                       (li "True tail calls")
                       (li "Delimited continuations")
                       (li "Concurrency built on JS promises"))))
                 (blockquote
             "If Guile’s web bootie doesn’t pack all the funk in its trunk, probably
it’s just junk."
             (footer
              (cite "— Andy Wingo")))))))
  (slide "Guile"
         `((div (@ (class "row"))
                (ul
                 (li "Spritely’s Scheme implementation of choice")
                 (li "Hoot’s " (em "only") " dependency!")
                 (li "Readily available in all Linux distros"))
                ,(img "static/kid-programming-a-robot.png"
                      '((class "white-background"))))))
  (slide "Takeaways"
         `((ul
            (li "Wasm GC opens doors for dynamic languages on the web")
            (li "Wasm is an escape hatch from npm hell")
            (li "Hoot is a good foundation for self-hosted apps"))))
  (slide "Demo: Strigoform"
         '((iframe (@ (frameborder "0")
                      (src "https://itch.io/embed-upload/8997570?color=333333")
                      (width "724")
                      (height "984"))
                   (a (@ (href "https://davexunit.itch.io/strigoform"))
                      "Play Strigoform on itch.io"))))
  (slide "Demo: Cirkoban"
         '((iframe (@ (frameborder "0")
                      (src "https://itch.io/embed-upload/10583669?color=222034")
                      (width "1284")
                      (height "984"))
                   (a (@ (href "https://davexunit.itch.io/cirkoban"))
                      "Play Cirkoban on itch.io"))))
  (slide "Demo: This talk"
         `((div (@ (class "row"))
                (p "This slideshow is a Hoot application!")
                ,(img "static/hoot-wasm-slideshow-source.png"))))
  (slide "Our supporters"
         `(,(img "static/supporters.png")))
  (slide "Thanks! Questions?"
         `(,(img "static/2020-10-31_spritely_scene.jpg"
                 '((style "width: 60%;")))
           (p "Learn more about Spritely:")
           (p (a (@ (href "https://spritely.institute"))
                 "https://spritely.institute"))
           (p "Support us with a donation:")
           (p (a (@ (href "https://every.org/spritely"))
                 "https://every.org/spritely")))
         #:title? #t)))
